#include <stdbool.h>
#include <avr/io.h>
#include <stdlib.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>

#include "waveforms.h"
#include "notes.h"

#define SAMPLE_RATE (44100UL)
#define MIDI_BAUD (31250UL)

volatile int8_t output;

volatile uint8_t current_note = 0;
volatile int16_t current_pitch_bend = 0;
volatile uint8_t current_velocity = 0;

volatile uint8_t envelope_m_state = 0;
volatile uint16_t envelope_m_attack = 30;
volatile uint16_t envelope_m_decay = 5;
volatile uint16_t envelope_m_sustain = 240 << 8;
volatile uint16_t envelope_m_release = 40;
volatile uint16_t envelope_m_value = 0;
volatile uint16_t envelope_m_amplitude = 256;

volatile uint8_t wave_m = 0;
volatile uint16_t phase_m = 0;
volatile uint16_t phase_m_delta = 196;
volatile uint16_t phase_m_scale = 515;

volatile uint8_t envelope_c_state = 0;
volatile uint16_t envelope_c_attack = 400;
volatile uint16_t envelope_c_decay = 100;
volatile uint16_t envelope_c_sustain = 96 << 8;
volatile uint16_t envelope_c_release = 30;
volatile uint16_t envelope_c_value = 0;
volatile uint16_t envelope_c_amplitude = 256;

volatile uint8_t wave_c = 0;
volatile uint16_t phase_c = 0;
volatile uint16_t phase_c_delta = 389;
volatile uint16_t phase_c_scale = 1024;


volatile uint8_t envelope_advance = 0;

ISR(TIMER1_COMPA_vect) {
	
	// Copy the previously-generated value to the output.
	OCR0A = output + 128;
	OCR0B = output + 128;
	
	if (envelope_advance++ % 8 == 0) {
	
		switch (envelope_m_state) {
			case 0: // idle
				envelope_m_value = 0;
				phase_m = 0;
				break;
			case 1: // attack
				if (envelope_m_attack >= ((uint16_t)(0xFFFF - envelope_m_value))) {
					envelope_m_value = 0xFFFF;
					envelope_m_state = 2;
				} else {
					envelope_m_value += envelope_m_attack;					
				}
				break;
			case 2: // decay
				if (envelope_m_value > envelope_m_decay) {
					envelope_m_value -= envelope_m_decay;
				} else {
					envelope_m_value = envelope_m_sustain;
				}
				if (envelope_m_value <= envelope_m_sustain) {
					envelope_m_value = envelope_m_sustain;
					envelope_m_state = 3;
				}
				break;
			case 3: // sustain
				envelope_m_value = envelope_m_sustain;
				break;
			case 4: // release
				if (envelope_m_release >= envelope_m_value) {
					envelope_m_value = 0;
					envelope_m_state = 0;
				} else if (envelope_m_value > envelope_m_sustain && envelope_m_value > envelope_m_decay) {
					envelope_m_value -= envelope_m_decay;
				} else {
					envelope_m_value -= envelope_m_release;
				}
				break;
		}
		
		switch (envelope_c_state) {
			case 0: // idle
				envelope_c_value = 0;
				phase_c = 0;
				break;
			case 1: // attack
				if (envelope_c_attack >= ((uint16_t)(0xFFFF - envelope_c_value))) {
					envelope_c_value = 0xFFFF;
					envelope_c_state = 2;
				} else {
					envelope_c_value += envelope_c_attack;					
				}
				break;
			case 2: // decay
				if (envelope_c_value > envelope_c_decay) {
					envelope_c_value -= envelope_c_decay;
				} else {
					envelope_c_value = envelope_c_sustain;
				}
				if (envelope_c_value <= envelope_c_sustain) {
					envelope_c_value = envelope_c_sustain;
					envelope_c_state = 3;
				}
				break;
			case 3: // sustain
				envelope_c_value = envelope_c_sustain;
				break;
			case 4: // release
				if (envelope_c_release >= envelope_c_value) {
					envelope_c_value = 0;
					envelope_c_state = 0;
				} else if (envelope_c_value > envelope_c_sustain && envelope_c_value > envelope_c_decay) {
					envelope_c_value -= envelope_c_decay;
				} else {
					envelope_c_value -= envelope_c_release;
				}
				break;
		}
	}
	
	// Advance the modulator phase
	phase_m += phase_m_delta;
	int8_t mv = pgm_read_byte(&waveforms[(wave_m << 8) + (phase_m >> 8)]);
	int16_t cpo = mv * (int16_t)(envelope_m_value >> 8);
	
	if (envelope_m_amplitude != 256) {
		cpo = ((int32_t)cpo * (uint32_t)envelope_c_amplitude) / 256;
	}
	
	// Advance the carrier phase
	phase_c += phase_c_delta;
	int8_t cv = pgm_read_byte(&waveforms[(wave_c << 8) + ((uint16_t)(phase_c + cpo) >> 8)]);
	
	cv = (cv * (envelope_c_value >> 8)) >> 8;
	
	if (envelope_c_amplitude != 256) {
		cv = ((int32_t)cv * ((uint32_t)envelope_c_amplitude)) / 256;
	}
	
	if (current_velocity < 255) {
		cv = (cv * (current_velocity + 1)) >> 8;
	}
	
	output = cv;
	
}

volatile uint8_t midi_status;
volatile uint8_t midi_data[2];
volatile uint8_t midi_data_index;
volatile uint8_t midi_data_bytes_per_message;

#define MIDI_QUEUE_LENGTH (128)
volatile uint8_t midi_queue[3 * MIDI_QUEUE_LENGTH];
volatile uint8_t midi_queue_read = 0;
volatile uint8_t midi_queue_write = 0;
volatile uint8_t midi_queue_count = 0;

ISR(USART_RX_vect) {
	uint8_t data = UDR0;
	if (data >= 0xF8) {
		// system control message, ignore
	} else if (data >= 0x80) {
		midi_status = data;
		midi_data_index = 0;
		midi_data[0] = 0;
		midi_data[1] = 0;
		midi_data_bytes_per_message = 2;
		switch (midi_status & 0xF0) {
			case 0xC0: // program change
			case 0xD0: // channel pressure (after-touch)
				midi_data_bytes_per_message = 1;
				break;
			case 0xF0: // sysex, time code, song position, song select, tune request, end of exclusive
				midi_data_bytes_per_message = 0;
				break;
		}
	} else if (midi_data_bytes_per_message > 0) {
		
		midi_data[midi_data_index++] = data;
		
		if (midi_data_index == midi_data_bytes_per_message) {
			
			if (midi_queue_count < MIDI_QUEUE_LENGTH) {
				midi_queue[midi_queue_write++] = midi_status;
				midi_queue[midi_queue_write++] = midi_data[0];
				midi_queue[midi_queue_write++] = midi_data[1];
				
				if (midi_queue_write >= 3 * MIDI_QUEUE_LENGTH) {
					midi_queue_write = 0;
				}
				++midi_queue_count;
			}
			
			midi_data_index = 0;
			midi_data[0] = 0;
			midi_data[1] = 0;
		}
	}
}

void pitch_bend(uint8_t note, int16_t amount) {
	
	current_pitch_bend = amount;
	
	// amount goes from -8192 to +8191
	uint16_t phase_increment = pgm_read_dword(&notes[note]);
	
	uint32_t scale_factor_c = phase_c_scale;
	uint32_t scale_factor_m = phase_m_scale;
	
	amount /= 8;

	if (amount > 0) {
		scale_factor_c *= (amount + 0x2000);
		scale_factor_m *= (amount + 0x2000);
		scale_factor_c >>= 13;
		scale_factor_m >>= 13;
	} else if (amount < 0) {
		scale_factor_c <<= 13;
		scale_factor_m <<= 13;
		scale_factor_c /= (0x2000 - amount);
		scale_factor_m /= (0x2000 - amount);
	}
	
	uint16_t phase_increment_c = (phase_increment * scale_factor_c) >> 10;
	uint16_t phase_increment_m = (phase_increment * scale_factor_m) >> 10;
	
	cli();
	phase_c_delta = phase_increment_c;
	phase_m_delta = phase_increment_m;
	sei();
}

void note_on(uint8_t note, uint8_t velocity) {
	current_note = note;
	cli();
	envelope_m_state = 1;
	//envelope_m_value = 0;
	envelope_c_state = 1;
	//envelope_c_value = 0;
	current_velocity = (velocity << 1) | (velocity >> 7);
	sei();
	pitch_bend(note, current_pitch_bend);
}

void note_off(uint8_t note) {
	if (note == current_note) {
		cli();
		envelope_m_state = 4;
		envelope_c_state = 4;
		sei();
	}
}

void change_program(uint8_t program) {
	
	phase_c_scale = 1024;
	phase_m_scale = 1024;
	
	envelope_c_amplitude = 256;
	envelope_m_amplitude = 256;
	
	wave_c = 0;
	wave_m = 0;
	
	switch (program) {
		case 0: // brass
			

			envelope_c_attack = 400;
			envelope_c_decay = 100;
			envelope_c_sustain = 96 << 8;
			envelope_c_release = 30;
			
			phase_m_scale = 515;

			envelope_m_attack = 20;
			envelope_m_decay = 5;
			envelope_m_sustain = 240 << 8;
			envelope_m_release = 40;
			
			envelope_m_amplitude = 320;

			break;
		case 1: // flute

			envelope_c_attack = 118;
			envelope_c_decay = 23;
			envelope_c_sustain = 128 << 8;
			envelope_c_release = 23;

			phase_m_scale = 2068;

			envelope_m_attack = 23;
			envelope_m_decay = 30;
			envelope_m_sustain = 64 << 8;
			envelope_m_release = 40;
			envelope_m_amplitude = 100;
		
			break;
			
		case 2: // electric piano

			envelope_c_attack = 5292;
			envelope_c_decay = 25;
			envelope_c_sustain = 64 << 8;
			envelope_c_release = 5;

			phase_m_scale = 3072;

			envelope_m_attack = 5292;
			envelope_m_decay = 49;
			envelope_m_sustain = 128 << 8;
			envelope_m_release = 81;
		
			break;
		
		case 3: // organ
		
		
			envelope_c_attack = 65535;
			envelope_c_decay = 10;
			envelope_c_sustain = 200 << 8;
			envelope_c_release = 48;

			phase_m_scale = 512;
			envelope_m_attack = 1200;
			envelope_m_decay = 200;
			envelope_m_sustain = 60 << 8;
			envelope_m_release = 48;
			wave_m = 2;
			
			
			break;
			
		case 4: // acoustic guitar
		
			phase_c_scale = 2048;
			envelope_c_attack = 65535;
			envelope_c_decay = 40;
			envelope_c_sustain = 0;
			envelope_c_release = 20;

			phase_m_scale = 512;
			envelope_m_attack = 600;
			envelope_m_decay = 100;
			envelope_m_sustain = 200 << 8;
			envelope_m_release = 48;
			wave_m = 2;
		
		
			break;
			
		case 5: // electric guitar
		
			phase_c_scale = 1536;
			envelope_c_attack = 65535;
			envelope_c_decay = 18;
			envelope_c_sustain = 0;
			envelope_c_release = 22;
			wave_c = 2;

			phase_m_scale = 512;
			envelope_m_attack = 600;
			envelope_m_decay = 5000;
			envelope_m_sustain = 200 << 8;
			envelope_m_release = 48;
			wave_m = 2;
		
			break;
	}
}


int main(void) {
	
	PORTC |= _BV(5);
	
	// We'll use fast PWM mode to generate an output analogue waveform.
	TCCR0A = (1 << COM0A1) | (0 << COM0A0) | (1 << COM0B1) | (0 << COM0B0) | (1 << WGM01) | (1 << WGM00);
	TCCR0B = (0 << WGM02) | (1 << CS00); // No prescaler
	DDRD |= _BV(6) | _BV(5);

	// Set up timer to run at the output sample rate.
	TCCR1B |= (0b001 << CS10); // Run at native clock speed, no prescaling.
	TIMSK1 |= _BV(OCIE1A);     // Interrupt on match A.
	TCCR1B |= _BV(WGM12);      // Mode 4 (TOP=OCR1A, CTC).
	OCR1A = (F_CPU / SAMPLE_RATE) - 1;
	
	// Set up the USART to receive MIDI
	UCSR0B = (1 << RXCIE0) | (1 << RXEN0) | (0 << UCSZ02);
	UCSR0C = (1 << UCSZ01) | (1 << UCSZ00); // 8 bits per character
	UBRR0 = (F_CPU / 16 / MIDI_BAUD) - 1;
	
	sei();
	
	uint8_t old_pin_c = PINC;
	for (;;) {
		
		while (midi_queue_count > 0) {
			uint8_t status = midi_queue[midi_queue_read++];
			uint8_t data0 = midi_queue[midi_queue_read++];
			uint8_t data1 = midi_queue[midi_queue_read++];
			if (midi_queue_read >= 3 * MIDI_QUEUE_LENGTH) {
				midi_queue_read = 0;
			}
			--midi_queue_count;
			switch (status & 0xF0) {
				case 0x80: // note off, note, velocity
					note_off(data0);
					break;
				case 0x90: // note on, note, velocity
					if (data1 == 0) {
						note_off(data0);
					} else {
						note_on(data0, data1);
					}
					break;
				case 0xA0: // poly pressure (aftertouch), note, pressure
					break;
				case 0xB0: // control change, control, value
					break;
				case 0xC0: // program change, program
					change_program(data0);
					break;
				case 0xD0: // channel pressure (aftertouch), pressure
					break;
				case 0xE0: // pitch bend, lsb, msb
					pitch_bend(current_note, (data0 | (data1 << 7)) - 0x2000);
					break;
			}
		}
		
		uint8_t new_pin_c = PINC;
		uint8_t changed_pin_c = old_pin_c ^ new_pin_c;
		if (changed_pin_c & _BV(5)) {
			if (new_pin_c & _BV(5)) {
				
			} else {
				
			}
		}
		old_pin_c = new_pin_c;
	}
}